# -*- mode: python; coding: utf-8 -*-
#
# PySysInfo
#
# sysinfo.py
#
# Copyright (C) 2003-2019 Elmar Hoffmann
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import sys
import os
import time
import socket
import re


class SysInfoBase:
    "private class, use SysInfo"

    def _ReadFile(self, Filename):
        try:
            f = file(Filename)
            buf = f.read()
            f.close()
        except IOError:
            return False
        return buf

    def _ReadFileLine(self, Filename):
        try:
            f = file(Filename)
            line = f.readline()
            f.close()
        except IOError:
            return False
        return line

    def _ReadFileLines(self, Filename):
        try:
            f = file(Filename)
            lines = f.readlines()
            f.close()
        except IOError:
            return False
        return lines

    def _ReadCommand(self, Command, Args, StdErr=False):
        Args.insert(0, Command)
        buf = False
        try:
            pout = os.pipe()
        except OSError:
            return False
        try:
            fdout = os.fdopen(pout[0], "r")
        except OSError:
            os.close(pout[0])
            os.close(pout[1])
            return False
        try:
            pid = os.fork()
        except OSError:
            fdout.close()
            os.close(pout[1])
            return False
        else:
            if pid == 0:
                # child
                try:
                    os.close(pout[0])
                    os.dup2(pout[1], sys.stdout.fileno())
                    if StdErr:
                        os.dup2(pout[1], sys.stderr.fileno())
                    os.close(pout[1])

                    os.execv(Command, Args)
                except:
                    pass
                os._exit(127)
            else:
                # parent
                os.close(pout[1])
                buf = fdout.read()
                while True:
                    wpid, status = os.waitpid(pid, 0)
                    if os.WIFSTOPPED(status):
                        continue
                    else:
                        break
                fdout.close()
        return buf

    def _SecToTime(self, sec):
        sec = int(sec)
        day = sec / 86400
        sec %= 86400
        hour = sec / 3600
        sec %= 3600
        min = sec / 60
        sec %= 60
        return (day, hour, min, sec)
                                    
    def BytesToHuman(self, bytes):
        if bytes >= 1024:
            x = bytes / 1024.0
            if x >= 1024:
                x /= 1024
                if x >= 1024:
                    x /= 1024
                    if x >= 1024:
                        x /= 1024
                        res = "%.2f TB" % x
                    else:
                        res = "%.2f GB" % x
                else:
                    res = "%.2f MB" % x
            else:
                res = "%.2f KB" % x
        else:
            res = "%d B" % bytes
        return res

    def VHostname(self):
        try:
            return os.environ['SERVER_NAME']
        except KeyError:
            return "N/A"

    def CanonicHostname(self):
        "get canonical hostname"
        return socket.gethostbyaddr(socket.gethostname())[0]

    def IPAddress(self):
        try:
            return os.environ['SERVER_ADDR']
        except KeyError:
            return "N/A"

    def KernelName(self):
        "get kernel name"
        return "N/A"
                    
    def KernelVersion(self):
        "get kernel version"
        return "N/A"

    def Uptime(self):
        """get system uptime
        returns tuple (days, hours, minutes, seconds)"""
        return (0, 0, 0, 0)

    def LoadAvg(self):
        """get system load average
        returns tuple (1min, 5min, 15min)"""
        return (0, 0, 0)

    def Users(self):
        "get number of logged in users"
        lines = self._ReadCommand("/usr/bin/who", ["-q"])
        if lines:
            return re.search("^# users=(\d+)", lines, re.MULTILINE).groups()[0]
        return 0

    def CPU(self):
        """get CPU information
        returns list of dictionaries"""
        return []

    def PCI(self):
        """get list of PCI cards
        returns list of strings"""
        return []

    def SCSI(self):
        """get list of SCSI devices
        returns list of dictionaries"""
        return []

    def IDE(self, Human=False):
        """get list of IDE devices
        returns list of dictionaries"""
        return []

    def Interfaces(self, Human=False):
        """get list of network interfaces
        returns list of dictionaries"""
        return []

    def Memory(self, Human=False):
        """get memory information
        returns dictionary"""
        return {}

    def Filesystems(self, HaveType, Human=False):
        """get mounted filesystems information
        returns list of dictionaries"""
        fss = []
        args = "-k"
        mountpoint_exclude = (
            "/dev/shm",
            "/run/",
            "/sys/",
        )
        type_exclude = (
            "devtmpfs",
            )
        if HaveType:
            args += "T"
        lines = self._ReadCommand("/bin/df", [args])
        if lines:
            fs = re.compile("^(?P<device>[^\s]+)\s+((?P<type>[^\s]+)\s+)?(?P<total>\d+)\s+(?P<used>\d+)\s+(?P<free>\d+)\s+(?P<percused>\d+)%\s+(?P<mountpoint>.*)$", re.MULTILINE)

            iter = fs.finditer(lines)
            while True:
                try:
                    fsd = iter.next().groupdict()
                except StopIteration:
                    break
                else:
                    try:
                        if fsd['type'] in type_exclude:
                            raise ValueError
                        for exclude in mountpoint_exclude:
                            if exclude[-1] == '/':
                                if fsd['mountpoint'] == exclude[:-1] \
                                   or fsd['mountpoint'].startswith(exclude):
                                    raise ValueError
                            else:
                                if fsd['mountpoint'] == exclude:
                                    raise ValueError
                        fsd['total'] = long(fsd['total'])
                        fsd['used'] = long(fsd['used'])
                        fsd['free'] = long(fsd['free'])
                        fsd['percused'] = long(fsd['percused'])
                    except ValueError:
                        continue
                    if Human:
                        fsd['total'] = self.BytesToHuman(fsd['total'] * 1024)
                        fsd['used'] = self.BytesToHuman(fsd['used'] * 1024)
                        fsd['free'] = self.BytesToHuman(fsd['free'] * 1024)
                    fss.append(fsd)
        return fss


class SysInfoLinux(SysInfoBase):
    "private class, use SysInfo"

    def KernelName(self):
        "get kernel name"
        return "Linux"

    def KernelVersion(self):
        "get kernel version"
        line = self._ReadFileLine("/proc/version")
        if line:
            return line.split()[2]
        else:
            return "N/A"

    def Uptime(self):
        """get system uptime
        returns tuple (days, hours, minutes, seconds)"""
        line = self._ReadFileLine("/proc/uptime")
        if line:
            return self._SecToTime(float(line.split()[0]))
        else:
            return (0, 0, 0, 0)

    def LoadAvg(self):
        """get system load average
        returns tuple (1min, 5min, 15min)"""
        line = self._ReadFileLine("/proc/loadavg")
        if line:
            return tuple(line.split()[0:3])
        else:
            return (0, 0, 0)

    def CPU(self):
        """get CPU information
        returns list of dictionaries"""
        cpus = []
        cpu = {}
        lines = self._ReadFileLines("/proc/cpuinfo")
        if lines:
            for line in lines:
                line = line.split(":")
                if len(line) < 2:
                    continue
                line = map(str.strip, line)
                line[0] = line[0].lower()
                # Intel
                if (line[0] == "processor") and (int(line[1]) > 0):
                    cpus.append(cpu)
                    cpu = {}
                # Intel / Alpha, PPC
                if (line[0] == "model name") or (line[0] == "cpu model") or (line[0] == "cpu"):
                    cpu['model'] = line[1]
                    continue
                # Intel / PPC
                if (line[0] == "cpu mhz") or (line[0] == "clock"):
                    cpu['mhz'] = line[1]
                    continue
                # Alpha
                if line[0] == "cycle frequency [hz]":
                    cpu['mhz'] = str(long(line[1])/1000000)
                    continue
                # Intel
                if line[0] == "cache size":
                    cpu['cache'] = line[1]
                    continue
                if line[0] == "bogomips":
                    cpu['bogomips'] = line[1]
                    continue
                # Alpha
                if line[0] == "cpus active":
                    cpu['count'] = int(line[1])
                    continue
            cpus.append(cpu)
            return cpus
        else:
            return []

    def PCI(self):
        """get list of PCI cards
        returns list of dictionaries"""
        cards = []
        lines = self._ReadCommand("/usr/bin/lspci", ["-m","-v"])
        if lines:
            card = re.compile("Class:\s+(?P<class>.*)\nVendor:\s+(?P<vendor>.*)\nDevice:\s+(?P<device>.*)")
            iter = card.finditer(lines)
            while True:
                try:
                    cardd = iter.next().groupdict()
                    class_ = cardd['class'].lower()
                    if (class_.find("bridge") < 0) and (class_.find("usb") < 0):
                        cards.append(cardd)
                except StopIteration:
                    break
        return cards

    def SCSI(self):
        """get list of SCSI devices
        returns list of dictionaries"""
        devs = []
        lines = self._ReadCommand("/usr/bin/lsscsi", ["--classic"])
        if lines:
            dev = re.compile("Vendor:\s+(?P<vendor>.*?)\s+Model:\s+(?P<model>.*?)\s+Rev:\s+(?P<rev>.*?)(\s+)?\n\s+Type:\s+(?P<type>.*?)\s+ANSI SCSI\s")
            iter = dev.finditer(lines)
            while True:
                try:
                    devs.append(iter.next().groupdict())
                except StopIteration:
                    break
        return devs

    def IDE(self, Human=False):
        """get list of IDE devices
        returns list of dictionaries"""
        try:
            files = os.listdir("/proc/ide")
        except OSError:
            return []
        devs = []
        for file in files:
            if file[:2] == "hd":
                dev = {}
                line = self._ReadFileLine(os.path.join("/proc/ide", file, "model"))
                if line:
                    line = line.strip()
                    dev['model'] = line
                line = self._ReadFileLine(os.path.join("/proc/ide", file, "media"))
                if line:
                    line = line.strip().lower()
                    if line.find("disk") >= 0:
                        dev['type'] = "Disk";
                    elif (line.find("cdrom") >= 0) or (line.find("cd-rom") >= 0):
                        dev['type'] = "CD-ROM";
                if  dev['type'] != "CD-ROM":
                    line = self._ReadFileLine(os.path.join("/proc/ide", file, "capacity"))
                    if line:
                        line = line.strip()
                        try:
                            capacity = long(line) * 512
                        except ValueError:
                            pass
                        else:
                            if Human:
                                dev['capacity'] = self.BytesToHuman(capacity)
                            else:
                                dev['capacity'] = capacity
                devs.append(dev)
        return devs

    def Interfaces(self, Human=False):
        """get list of network interfaces
        returns list of dictionaries"""
        ifs = []
        lines = self._ReadFile("/proc/net/dev")
        if lines:
            iface = re.compile("\s*(?P<name>[^\s:]+):\s*(?P<rx_bytes>\d+)\s+(?P<rx_packets>\d+)\s+(?P<rx_errs>\d+)\s+(?P<rx_drop>\d+)\s+\d+\s+\d+\s+\d+\s+\d+\s+(?P<tx_bytes>\d+)\s+(?P<tx_packets>\d+)\s+(?P<tx_errs>\d+)\s+(?P<tx_drop>\d+)\s+\d+\s+(?P<coll>\d+)", re.MULTILINE)
            iter = iface.finditer(lines)
            while True:
                try:
                    ifdict = iter.next().groupdict()
                except StopIteration:
                    break
                else:
                    for k, v in ifdict.items():
                        try:
                            ifdict[k] = long(ifdict[k])
                        except ValueError:
                            pass

                    ifdict['errs'] = ifdict['rx_errs'] + ifdict['tx_errs']
                    ifdict['drop'] = ifdict['rx_drop'] + ifdict['tx_drop']

                    if Human:
                        ifdict['rx_bytes'] = self.BytesToHuman(ifdict['rx_bytes'])
                        ifdict['tx_bytes'] = self.BytesToHuman(ifdict['tx_bytes'])

                    ifs.append(ifdict)
        return ifs

    def Memory(self, Human=False):
        """get memory information
        returns dictionary"""
        lines = self._ReadCommand("/usr/bin/free", ["-b"])
        if lines:
            h = re.search("\s*total\s+used\s+free\s+shared\s+(?P<head_buffers>[^\s]+)\s+(?P<head_cached>[^\s]+)", lines)
            m = re.search("Mem:\s*(?P<mem_total>\d+)\s+(?P<mem_used>\d+)\s+(?P<mem_free>\d+)\s+(?P<mem_shared>\d+)\s+(?P<mem_buffers>\d+)\s+(?P<mem_cached>\d+)", lines)
            r = re.search("-/\+ buffers/cache:\s*(?P<mem_realused>\d+)\s+(?P<mem_realfree>\d+)", lines)
            s = re.search("Swap:\s*(?P<swap_total>\d+)\s+(?P<swap_used>\d+)\s+(?P<swap_free>\d+)", lines)

            head = h.groupdict()
            mem = m.groupdict()
            mem.update(s.groupdict())

            if head['head_buffers'] == 'buffers' \
               and head['head_cached'] == 'cached':
                # classic free(1) output

                # "+/- buffers/cache" line
                mem.update(r.groupdict())
            elif head['head_buffers'] == 'buff/cache' \
                 and head['head_cached'] == 'available':
                # procps 3.3.10 free(1) output

                # "available" has taken the position of "cached"
                mem['mem_realfree'] = mem['mem_cached']
                del mem['mem_cached']

                # "used" is given with buffers and cache already excluded
                mem['mem_realused'] = mem['mem_used']
                mem['mem_used'] += mem['mem_buffers']

            for k, v in mem.items():
                try:
                    mem[k] = long(mem[k])
                except ValueError:
                    del mem[k]

            try:
                mem_percused = mem['mem_realused'] * 100 / mem['mem_total']
            except (KeyError, ZeroDivisionError):
                mem_percused = 0

            try:
                swap_percused = mem['swap_used'] * 100 / mem['swap_total']
            except (KeyError, ZeroDivisionError):
                swap_percused = 0

            if Human:
                for k, v in mem.items():
                    mem[k] = self.BytesToHuman(mem[k])

            mem['mem_percused'] = mem_percused
            mem['swap_percused'] = swap_percused

            return mem
        return {}

    def Filesystems(self, Human=False):
        """get mounted filesystems information
        returns list of dictionaries"""
        return SysInfoBase.Filesystems(self, True, Human)


class SysInfoBSD(SysInfoBase):
    "private class, use SysInfo"

    def _ReadSysCtl(self, key):
        return self._ReadCommand("/usr/sbin/sysctl", ["-n", key])

    def KernelName(self):
        "get kernel name"
        return "BSD"

    def KernelVersion(self):
        "get kernel version"
        lines = self._ReadSysCtl("kern.version")
        if lines:
            return lines.split()[1]
        else:
            return "N/A"

    def Uptime(self):
        """get system uptime
        returns tuple (days, hours, minutes, seconds)"""
        lines = self._ReadSysCtl("kern.boottime")
        if lines:
            return self._SecToTime(time.time() - float(lines.split()[0]))
        else:
            return (0, 0, 0, 0)

    def LoadAvg(self):
        """get system load average
        returns tuple (1min, 5min, 15min)"""
        lines = self._ReadSysCtl("vm.loadavg")
        if lines:
            return tuple(lines.split()[0:3])
        else:
            return (0, 0, 0)

    def CPU(self):
        """get CPU information
        returns list of dictionaries"""
        cpus = []
        cpu = {}
        cpu['model'] = self._ReadSysCtl("hw.model")
        cpu['count'] = self._ReadSysCtl("hw.ncpu")
        cpus.append(cpu)
        return cpus

    def PCI(self):
        """get list of PCI cards
        returns list of strings"""
        cards = []
        return cards

    def SCSI(self):
        """get list of SCSI devices
        returns list of dictionaries"""
        devs = []
        return devs

    def Interfaces(self, Human=False):
        """get list of network interfaces
        returns list of dictionaries"""
        ifs = []
        return ifs

    def Memory(self, Human=False):
        """get memory information
        returns dictionary"""
        return {}

    def Filesystems(self, Human=False):
        """get mounted filesystems information
        returns list of dictionaries"""
        return SysInfoBase.Filesystems(self, False, Human)


class SysInfoNetBSD(SysInfoBSD):
    "private class, use SysInfo"

    def KernelName(self):
        "get kernel name"
        return "NetBSD"


class SysInfoOpenBSD(SysInfoBSD):
    "private class, use SysInfo"

    def KernelName(self):
        "get kernel name"
        return "OpenBSD"



if sys.platform[:5] == "linux":
    class SysInfo(SysInfoLinux):
        pass
elif sys.platform[:6] == "netbsd":
    class SysInfo(SysInfoNetBSD):
        pass
elif sys.platform[:7] == "openbsd":
    class SysInfo(SysInfoOpenBSD):
        pass
